@@ -10,6 +10,7 @@
 #include <Xm/Form.h>
 #include <Xm/RowColumn.h>
 #include <Xm/PushB.h>
+#include <Xm/Label.h>
 #include <Xm/CascadeB.h>
 #include <Xm/ScrollBar.h>
 #include <Xm/MenuShell.h>
@@ -30,10 +31,17 @@ static Widget textArea;
 static Widget menuBar;
 
 static void scroll_cb __ARGS((Widget w, XtPointer client_data, XtPointer call_data));
-static XmString MenuStringCreate __ARGS((char_u *text, int *mnemonic));
+static void gui_motif_add_actext __ARGS((GuiMenu *menu));
+static void gui_mch_recurse_tearoffs __ARGS((GuiMenu *menu, int val));
 static void gui_mch_compute_menu_height __ARGS((Widget));
 static void gui_mch_submenu_colors __ARGS((GuiMenu *mp));
 
+#if (XmVersion >= 1002)
+# define STRING_TAG  XmFONTLIST_DEFAULT_TAG
+#else
+# define STRING_TAG  XmSTRING_DEFAULT_CHARSET
+#endif
+
 /*
  * Call-back routines.
  */
@@ -194,61 +202,78 @@ gui_mch_set_menu_pos(x, y, w, h)
 }
 
 /*
- * Create an XmString for a menu entry. If an ampersand (&) is found then
- * extract the following character as the menu mnemonic and remove the
- * ampersand.
+ * Enable or disable mnemonics for the toplevel menus.
  */
-    static XmString 
-MenuStringCreate(text, mnemonic) 
-    char_u	*text;
-    int		*mnemonic;
+    void
+gui_motif_set_mnemonics(enable)
+    int		enable;
 {
-    char_u	*menu_text;
-    XmString	label;
+    GuiMenu	*menu;
 
-    menu_text = gui_menu_text(text, mnemonic);
-    if (menu_text == NULL)
-	return (XmString)NULL;
-
-#if (XmVersion >= 1002)
-    label = XmStringCreate((char *)menu_text, XmFONTLIST_DEFAULT_TAG);
-#else
-    label = XmStringCreate((char *)menu_text, XmSTRING_DEFAULT_CHARSET);
-#endif
-
-    vim_free(menu_text);
-    return label;
+    for (menu = gui.root_menu; menu != NULL; menu = menu->next)
+	if (menu->id != (Widget)0)
+	    XtVaSetValues(menu->id,
+		    XmNmnemonic, enable ? menu->mnemonic : NUL,
+		    NULL);
 }
 
     void
-gui_mch_add_menu(menu, parent)
-    GuiMenu *menu;
-    GuiMenu *parent;
+gui_mch_add_menu(menu, parent, idx)
+    GuiMenu	*menu;
+    GuiMenu	*parent;
+    int		idx;
 {
 #if (XmVersion >= 1002)
     Widget	widget;
 #endif
-    int		mnemonic = 0;
     XmString	label;
     Widget	shell;
-    GuiMenu	*mp;
-    int		idx;
 
-    if (!gui_menubar_menu(menu->name))
+#ifdef MOTIF_POPUP
+    if (gui_popup_menu(menu->name))
+    {
+	shell = XtVaCreateWidget("subMenuShell",
+		xmMenuShellWidgetClass, vimShell,
+		XmNwidth, 1,
+		XmNheight, 1,
+		XmNforeground, gui.menu_fg_pixel,
+		XmNbackground, gui.menu_bg_pixel,
+		NULL);
+	menu->id = shell;
+	menu->submenu_id = XtVaCreateWidget("rowColumnMenu",
+		xmRowColumnWidgetClass, shell,
+		XmNrowColumnType, XmMENU_POPUP,
+		XmNpopupEnabled, False,
+		NULL);
+	return;
+    }
+#endif
+
+    if (!gui_menubar_menu(menu->name)
+	    || (parent != NULL && parent->submenu_id == (Widget)0))
 	return;
 
-    label = MenuStringCreate(menu->name, &mnemonic);
+    label = XmStringCreate((char *)menu->dname, STRING_TAG);
     if (label == NULL)
 	return;
     menu->id = XtVaCreateWidget("subMenu",
-	xmCascadeButtonWidgetClass,
-	(parent == NULL) ? menuBar : parent->submenu_id,
-	XmNlabelString, label,
-	XmNforeground, gui.menu_fg_pixel,
-	XmNbackground, gui.menu_bg_pixel,
-        XmNmnemonic, mnemonic,
-	NULL);
-    /* XtFree((char *)label); makes Lesstif crash */
+	    xmCascadeButtonWidgetClass,
+	    (parent == NULL) ? menuBar : parent->submenu_id,
+	    XmNlabelString, label,
+	    XmNforeground, gui.menu_fg_pixel,
+	    XmNbackground, gui.menu_bg_pixel,
+	    XmNmnemonic, p_wak[0] == 'n' ? NUL : menu->mnemonic,
+#if (XmVersion >= 1002)
+	    XmNpositionIndex, idx,
+#endif
+	    NULL);
+    XmStringFree(label);
+
+    if (menu->id == (Widget)0)		/* failed */
+	return;
+
+    /* add accelerator text */
+    gui_motif_add_actext(menu);
 
     shell = XtVaCreateWidget("subMenuShell",
 	xmMenuShellWidgetClass, menu->id,
@@ -265,9 +290,12 @@ gui_mch_add_menu(menu, parent)
 #endif
 	NULL);
 
+    if (menu->submenu_id == (Widget)0)		/* failed */
+	return;
+
 #if (XmVersion >= 1002)
     /* Set the colors for the tear off widget */
-    if ((widget = XmGetTearOffControl(menu->submenu_id)) != (Widget)NULL)
+    if ((widget = XmGetTearOffControl(menu->submenu_id)) != (Widget)0)
 	XtVaSetValues(widget,
 	    XmNforeground, gui.menu_fg_pixel,
 	    XmNbackground, gui.menu_bg_pixel,
@@ -287,37 +315,62 @@ gui_mch_add_menu(menu, parent)
 	    XmNmenuHelpWidget, menu->id,
 	    NULL);
 
+    /*
+     * When we add a top-level item to the menu bar, we can figure out how
+     * high the menu bar should be.
+     */
     if (parent == NULL)
-    {
-#if (XmVersion >= 1002)
-	/* XmNpositionIndex was added in Motif 1.2 */
-	/*
-	 * Put the menu at the right position.
-	 */
-	mp = gui.root_menu;
-	for (idx = 0; mp != NULL; ++idx)
-	{
-	    /*
-	     * Skip the Help menu.
-	     */
-	    if (STRCMP((char *)mp->dname, "Help") == 0)
-		--idx;
-	    else if (mp == menu)
-	    {
-		XtVaSetValues(menu->id,
-		    XmNpositionIndex, idx,
-		    NULL);
-		break;
-	    }
-	    mp = mp->next;
-	}
-#endif
-
-	/*
-	 * When we add a top-level item to the menu bar, we can figure out how
-	 * high the menu bar should be.
-	 */
 	gui_mch_compute_menu_height(menu->id);
+}
+
+/*
+ * Add mnemonic and accelerator text to a menu button.
+ */
+    static void
+gui_motif_add_actext(menu)
+    GuiMenu	*menu;
+{
+    XmString	label;
+
+    /* Add accelrator text, if there is one */
+    if (menu->actext != NULL && menu->id != (Widget)0)
+    {
+	label = XmStringCreate((char *)menu->actext, STRING_TAG);
+	if (label == NULL)
+	    return;
+	XtVaSetValues(menu->id, XmNacceleratorText, label, NULL);
+	XmStringFree(label);
+    }
+}
+
+    void
+gui_mch_toggle_tearoffs(enable)
+    int		enable;
+{
+    int		val;
+
+    if (enable)
+	val = (int)XmTEAR_OFF_ENABLED;
+    else
+	val = (int)XmTEAR_OFF_DISABLED;
+    XtVaSetValues(menuBar, XmNtearOffModel, val, NULL);
+    gui_mch_recurse_tearoffs(gui.root_menu, val);
+}
+
+    static void
+gui_mch_recurse_tearoffs(menu, val)
+    GuiMenu	*menu;
+    int		val;
+{
+    while (menu != NULL)
+    {
+	if (!gui_popup_menu(menu->name))
+	{
+	    if (menu->submenu_id != (Widget)0)
+		XtVaSetValues(menu->submenu_id, XmNtearOffModel, val, NULL);
+	    gui_mch_recurse_tearoffs(menu->children, val);
+	}
+	menu = menu->next;
     }
 }
 
@@ -397,28 +450,40 @@ gui_mch_compute_menu_height(id)
 
 
     void
-gui_mch_add_menu_item(menu, parent)
-    GuiMenu *menu;
-    GuiMenu *parent;
+gui_mch_add_menu_item(menu, parent, idx)
+    GuiMenu	*menu;
+    GuiMenu	*parent;
+    int		idx;
 {
-    int		mnemonic = 0;
     XmString	label;
 
-    label = MenuStringCreate(menu->name, &mnemonic);
+    /* No parent, must be a non-menubar menu */
+    if (parent->submenu_id == (Widget)0)
+	return;
+
+    label = XmStringCreate((char *)menu->dname, STRING_TAG);
     if (label == NULL)
 	return;
     menu->submenu_id = (Widget)0;
     menu->id = XtVaCreateWidget("subMenu",
-        xmPushButtonWidgetClass, parent->submenu_id,
-        XmNlabelString, label,
-        XmNforeground, gui.menu_fg_pixel,
-        XmNbackground, gui.menu_bg_pixel,
-        XmNmnemonic, mnemonic,
-        NULL);
-    /* XtFree((char *)label); makes Lesstif crash */
+	xmPushButtonWidgetClass, parent->submenu_id,
+	XmNlabelString, label,
+	XmNforeground, gui.menu_fg_pixel,
+	XmNbackground, gui.menu_bg_pixel,
+	XmNmnemonic, menu->mnemonic,
+#if (XmVersion >= 1002)
+	XmNpositionIndex, idx,
+#endif
+	NULL);
+    XmStringFree(label);
 
-    XtAddCallback(menu->id, XmNactivateCallback, gui_x11_menu_cb,
-	(XtPointer)menu);
+    if (menu->id != (Widget)0)
+    {
+	XtAddCallback(menu->id, XmNactivateCallback, gui_x11_menu_cb,
+		(XtPointer)menu);
+	/* add accelerator text */
+	gui_motif_add_actext(menu);
+    }
 }
 
     void
@@ -440,10 +505,11 @@ gui_mch_submenu_colors(mp)
 {
     while (mp != NULL)
     {
-	XtVaSetValues(mp->id,
-		XmNforeground, gui.menu_fg_pixel,
-		XmNbackground, gui.menu_bg_pixel,
-		NULL);
+	if (mp->id != (Widget)0)
+	    XtVaSetValues(mp->id,
+		    XmNforeground, gui.menu_fg_pixel,
+		    XmNbackground, gui.menu_bg_pixel,
+		    NULL);
 
 	if (mp->children != NULL)
 	{
@@ -451,7 +517,8 @@ gui_mch_submenu_colors(mp)
 	    Widget	widget;
 
 	    /* Set the colors for the tear off widget */
-	    if ((widget = XmGetTearOffControl(mp->submenu_id)) != (Widget)NULL)
+	    if (mp->submenu_id != (Widget)0
+	       && (widget = XmGetTearOffControl(mp->submenu_id)) != (Widget)0)
 		XtVaSetValues(widget,
 			XmNforeground, gui.menu_fg_pixel,
 			XmNbackground, gui.menu_bg_pixel,
@@ -500,7 +567,7 @@ gui_mch_destroy_menu(menu)
 	menu->id = (Widget)0;
 
 	if (parent == menuBar)
-	    gui_mch_compute_menu_height((Widget)NULL);
+	    gui_mch_compute_menu_height((Widget)0);
     }
 }
 
@@ -516,7 +583,8 @@ gui_mch_set_scrollbar_thumb(sb, val, size, max)
     int		    size;
     int		    max;
 {
-    XtVaSetValues(sb->id,
+    if (sb->id != (Widget)0)
+	XtVaSetValues(sb->id,
 		  XmNvalue, val,
 		  XmNsliderSize, size,
 		  XmNpageIncrement, (size > 2 ? size - 2 : 1),
@@ -532,12 +600,16 @@ gui_mch_set_scrollbar_pos(sb, x, y, w, h)
     int		    w;
     int		    h;
 {
-    XtVaSetValues(sb->id,
-		  XmNleftOffset, x,
-		  XmNtopOffset, y,
-		  XmNrightOffset, -x - w,
-		  XmNbottomOffset, -y - h,
-		  NULL);
+    if (sb->id != (Widget)0)
+    {
+	XtVaSetValues(sb->id,
+		      XmNleftOffset, x,
+		      XmNtopOffset, y,
+		      XmNrightOffset, -x - w,
+		      XmNbottomOffset, -y - h,
+		      NULL);
+	XtManageChild(sb->id);
+    }
 }
 
     void
@@ -559,18 +631,22 @@ gui_mch_create_scrollbar(sb, orient)
 	    XmNrightAttachment, XmATTACH_OPPOSITE_FORM,
 	    XmNbottomAttachment, XmATTACH_OPPOSITE_FORM,
 	    NULL);
-    XtAddCallback(sb->id, XmNvalueChangedCallback,
-		  scroll_cb, (XtPointer)sb->ident);
-    XtAddCallback(sb->id, XmNdragCallback,
-		  scroll_cb, (XtPointer)sb->ident);
-    XtAddEventHandler(sb->id, KeyPressMask, FALSE, gui_x11_key_hit_cb,
-	(XtPointer)0);
+    if (sb->id != (Widget)0)
+    {
+	XtAddCallback(sb->id, XmNvalueChangedCallback,
+		      scroll_cb, (XtPointer)sb->ident);
+	XtAddCallback(sb->id, XmNdragCallback,
+		      scroll_cb, (XtPointer)sb->ident);
+	XtAddEventHandler(sb->id, KeyPressMask, FALSE, gui_x11_key_hit_cb,
+	    (XtPointer)0);
+    }
 }
     void
 gui_mch_destroy_scrollbar(sb)
     GuiScrollbar    *sb;
 {
-    XtDestroyWidget(sb->id);
+    if (sb->id != (Widget)0)
+	XtDestroyWidget(sb->id);
 }
 
     void
@@ -595,9 +671,10 @@ gui_x11_get_wid()
     return(XtWindow(textArea));
 }
 
-#ifdef USE_BROWSE
+#if defined(USE_BROWSE) || defined(PROTO)
+
 /*
- * dialog related stuff 
+ * file selector related stuff
  */
 
 #include <Xm/FileSB.h>
@@ -609,11 +686,7 @@ typedef struct dialog_callback_arg
     int	    id;
 } dcbarg_t;
 
-#define DIALOG_FSELECT          300
-#define DIALOG_SAVEAS           301
-
-static Widget open_dialog;
-static Widget saveas_dialog;
+static Widget dialog_wgt;
 static char *browse_fname = NULL;
 static XmStringCharSet charset = (XmStringCharSet) XmSTRING_DEFAULT_CHARSET;
 				/* used to set up XmStrings */
@@ -622,42 +695,42 @@ static void DialogCancelCB __ARGS((Widget, XtPointer, XtPointer));
 static void DialogAcceptCB __ARGS((Widget, XtPointer, XtPointer));
 
 /*
- * Implementation of the :browse command
+ * Put up a file requester.
+ * Returns the selected name in allocated memory, or NULL for Cancel.
  */
 /* ARGSUSED */
     char_u *
 gui_mch_browse(saving, title, dflt, ext, initdir, filter)
-    int		saving;
-    char_u *	title;
-    char_u *	dflt;
-    char_u *	ext;
-    char_u *	initdir;
-    char_u *	filter;
+    int		saving;		/* select file to write */
+    char_u	*title;		/* title for the window */
+    char_u	*dflt;		/* default name */
+    char_u	*ext;		/* not used (extension added) */
+    char_u	*initdir;	/* initial directory, NULL for current dir */
+    char_u	*filter;	/* not used (file name filter) */
 {
-    Widget	which_w;
-    dcbarg_t	*arg;
+    XmString	dirstring;
+    char_u	dirbuf[MAXPATHL];
 
-    if ((arg = (dcbarg_t *)alloc((unsigned)sizeof(dcbarg_t))) == NULL)
-	return NULL;
+    dialog_wgt = XmCreateFileSelectionDialog(vimShell, (char *)title, NULL, 0);
 
-    if (saving == TRUE)
+    if (dflt == NULL)
+	dflt = (char_u *)"";
+    if (initdir == NULL || *initdir == NUL)
     {
-	saveas_dialog = XmCreateFileSelectionDialog(gui.root_menu->id,
-			"save as dialog", NULL, 0);
-	which_w = saveas_dialog;
-	arg->id = DIALOG_SAVEAS;
-    }
-    else
-    {
-	open_dialog = XmCreateFileSelectionDialog(gui.root_menu->id,
-			"file open dialog", NULL, 0);
-	which_w = open_dialog;
-	arg->id = DIALOG_FSELECT;
+	mch_dirname(dirbuf, MAXPATHL);
+	initdir = dirbuf;
     }
 
-    XtVaSetValues(which_w,
+    dirstring = XmStringCreate((char *)initdir, STRING_TAG);
+
+    /* How do we set the default file name to "dflt"? */
+    XtVaSetValues(dialog_wgt,
+	XmNdirectory,		dirstring,
+/*
+ These can cause a crash after ":hi Menu guifg=red".  Why?
 	XmNforeground,		gui.menu_fg_pixel,
 	XmNbackground,		gui.menu_bg_pixel,
+ */
 	XmNtroughColor,		gui.scroll_bg_pixel,
 /*
     currently, the background color of the input and selection
@@ -671,23 +744,29 @@ gui_mch_browse(saving, title, dflt, ext, initdir, filter)
 */
 	NULL);
 
-    XtAddCallback(which_w, XmNokCallback, DialogAcceptCB, (XtPointer)arg);
-    XtAddCallback(which_w, XmNcancelCallback, DialogCancelCB, (XtPointer)arg);
-    XtManageChild(which_w);
+    XtAddCallback(dialog_wgt, XmNokCallback, DialogAcceptCB, (XtPointer)0);
+    XtAddCallback(dialog_wgt, XmNcancelCallback, DialogCancelCB, (XtPointer)0);
+    XtManageChild(dialog_wgt);
 
     /* sit in a loop until the dialog box has gone away */
     do
     {
-	XtAppProcessEvent(XtWidgetToApplicationContext(which_w),
+	XtAppProcessEvent(XtWidgetToApplicationContext(dialog_wgt),
 	    (XtInputMask)XtIMAll);
-    } while (XtIsManaged(which_w));
+    } while (XtIsManaged(dialog_wgt));
 
-    return (char_u *)browse_fname;
+    XtDestroyWidget(dialog_wgt);
+    XmStringFree(dirstring);
+
+    if (browse_fname == NULL)
+	return NULL;
+    return vim_strsave((char_u *)browse_fname);
 }
 
 /*
- * the code below was originally taken and modified from
- * /usr/examples/motif/xmsamplers/xmeditor.c * on Digital Unix 4.0d
+ * The code below was originally taken from
+ *	/usr/examples/motif/xmsamplers/xmeditor.c
+ * on Digital Unix 4.0d, but heavily modified.
  */
 
 /*
@@ -695,34 +774,17 @@ gui_mch_browse(saving, title, dflt, ext, initdir, filter)
  */
 /* ARGSUSED */
     static void
-DialogCancelCB(w, client_data, call_data) 
+DialogCancelCB(w, client_data, call_data)
     Widget	w;		/*  widget id		*/
     XtPointer	client_data;	/*  data from application   */
     XtPointer	call_data;	/*  data from widget class  */
 {
-    dcbarg_t *arg = (dcbarg_t *)client_data;
-
-    switch (arg->id)
+    if (browse_fname != NULL)
     {
-	case DIALOG_FSELECT:
-	case DIALOG_SAVEAS:
-	    if (browse_fname != NULL)
-	    {
-		XtFree(browse_fname);
-		browse_fname = NULL;
-	    }
-	    if (arg->id == DIALOG_FSELECT)
-		XtUnmanageChild(open_dialog);
-	    else
-		XtUnmanageChild(saveas_dialog);
-	    XtFree(client_data);
-	    break;
-	default:
-	    /* a unknown client_data was recieved and
-	       there is no setup to handle this */
-	    /* fprintf(stderr, "Warning: in cancel callback\n"); */
-	    break;
+	XtFree(browse_fname);
+	browse_fname = NULL;
     }
+    XtUnmanageChild(dialog_wgt);
 }
 
 /*
@@ -730,62 +792,215 @@ DialogCancelCB(w, client_data, call_data)
  */
 /* ARGSUSED */
     static void
-DialogAcceptCB(w, client_data, call_data) 
+DialogAcceptCB(w, client_data, call_data)
     Widget	w;		/*  widget id		*/
     XtPointer	client_data;	/*  data from application   */
     XtPointer	call_data;	/*  data from widget class  */
 {
-    dcbarg_t *arg = (dcbarg_t *)client_data;
+    XmFileSelectionBoxCallbackStruct *fcb;
 
-    switch (arg->id)
+    if (browse_fname != NULL)
     {
-	case DIALOG_FSELECT:
-	case DIALOG_SAVEAS:
-	    /* open the file and read it into the text widget */
-	    if (browse_fname != NULL)
-	    {
-		XtFree(browse_fname);
-		browse_fname = NULL;
-	    }
-	    {
-		XmFileSelectionBoxCallbackStruct *fcb =
-		    (XmFileSelectionBoxCallbackStruct *) call_data;
-
-		/* get the filename from the file selection box */
-		XmStringGetLtoR(fcb->value, charset, &browse_fname);
-
-		/* popdown the file selection box */
-		if (arg->id == DIALOG_FSELECT)
-		{
-		    XtUnmanageChild(open_dialog);
-		    XtDestroyWidget(open_dialog);
-		}
-		else
-		{
-		    XtUnmanageChild(saveas_dialog);
-		    XtDestroyWidget(saveas_dialog);
-		}
-
-		XtFree(client_data);
-	    }
-	    break;
-	default:
-	    /* unknown callback type */
-	    /* fprintf(stderr, "Warning: in accept callback\n"); */
-	    break;
+	XtFree(browse_fname);
+	browse_fname = NULL;
     }
+    fcb = (XmFileSelectionBoxCallbackStruct *)call_data;
+
+    /* get the filename from the file selection box */
+    XmStringGetLtoR(fcb->value, charset, &browse_fname);
+
+    /* popdown the file selection box */
+    XtUnmanageChild(dialog_wgt);
 }
+
 #endif /* USE_BROWSE */
 
-#ifdef GUI_DIALOGUE
+#if defined(GUI_DIALOG) || defined(PROTO)
+
+static int	dialogStatus;
+
+static void butproc __ARGS((Widget w, XtPointer client_data, XtPointer call_data));
+
+/* ARGSUSED */
+    static void
+butproc(w, client_data, call_data)
+    Widget	w;
+    XtPointer	client_data;
+    XtPointer	call_data;
+{
+    dialogStatus = (int)(long)client_data + 1;
+}
+
+/* ARGSUSED */
     int
-gui_mch_dialogue(type, title, message, buttons, dfltbutton)
+gui_mch_dialog(type, title, message, buttons, dfltbutton)
     int		type;
     char_u	*title;
     char_u	*message;
     char_u	*buttons;
     int		dfltbutton;
 {
-    TODO
+    char_u		*buts;
+    char_u		*p, *next;
+    XtAppContext	app;
+    XEvent		event;
+    XmString		label;
+    int			butcount;
+    static Widget	dialogbb = NULL;
+    static Widget	dialogmessage = NULL;
+#define MAXBUT 10
+    Widget		dialogButton[MAXBUT];
+
+    if (title == NULL)
+	title = (char_u *)"Vim dialog";
+
+    /* if our pointer is currently hidden, then we should show it. */
+    gui_mch_mousehide(FALSE);
+
+    /* First time called: create the dialog and the message */
+    if (dialogbb == NULL)
+    {
+	dialogbb = XmCreateFormDialog(vimShell,
+		(char *)"dialog", NULL, 0);
+
+	dialogmessage = XtVaCreateManagedWidget("dialogMessage",
+		xmLabelWidgetClass, dialogbb,
+		XmNtopAttachment, XmATTACH_FORM,
+		XmNleftAttachment, XmATTACH_FORM,
+		XmNrightAttachment, XmATTACH_FORM,
+		NULL);
+    }
+
+    /* Set the message string */
+    label = XmStringLtoRCreate((char *)message, STRING_TAG);
+    if (label == NULL)
+	return -1;
+    XtVaSetValues(dialogmessage,
+	    XmNlabelString, label,
+	    NULL);
+    XmStringFree(label);
+
+    /* Set the title of the Dialog window */
+    label = XmStringCreateSimple((char *)title);
+    if (label == NULL)
+	return -1;
+    XtVaSetValues(dialogbb,
+	    XmNdialogTitle, label,
+	    XmNhorizontalSpacing, 20,
+	    XmNverticalSpacing, 20,
+	    NULL);
+    XmStringFree(label);
+
+    /* make a copy, so that we can insert NULs */
+    buts = vim_strsave(buttons);
+    if (buts == NULL)
+	return -1;
+
+    /*
+     * Create the buttons.
+     */
+    p = buts;
+    for (butcount = 0; butcount < MAXBUT; ++butcount)
+    {
+	for (next = p; *next; ++next)
+	{
+	    if (*next == DLG_HOTKEY_CHAR)
+		mch_memmove(next, next + 1, STRLEN(next));
+	    if (*next == DLG_BUTTON_SEP)
+	    {
+		*next++ = NUL;
+		break;
+	    }
+	}
+	label = XmStringCreate((char *)p, STRING_TAG);
+	if (label == NULL)
+	    break;
+
+	dialogButton[butcount] = XtVaCreateManagedWidget("button",
+		xmPushButtonWidgetClass, dialogbb,
+		XmNlabelString, label,
+		XmNtopAttachment, XmATTACH_WIDGET,
+		XmNtopWidget, dialogmessage,
+		NULL);
+	XmStringFree(label);
+	if (butcount == 0)
+	    XtVaSetValues(dialogButton[butcount],
+		    XmNleftAttachment, XmATTACH_FORM,
+		    NULL);
+	else
+	    XtVaSetValues(dialogButton[butcount],
+		    XmNleftAttachment, XmATTACH_WIDGET,
+		    XmNleftWidget, dialogButton[butcount - 1],
+		    NULL);
+
+	XtAddCallback(dialogButton[butcount], XmNactivateCallback,
+						butproc, (XtPointer)butcount);
+	if (*next == NUL)
+	    break;
+	p = next;
+    }
+    ++butcount;
+    vim_free(buts);
+
+    if (dfltbutton < 1)
+	dfltbutton = 1;
+    if (dfltbutton > butcount)
+	dfltbutton = butcount;
+    XtVaSetValues(dialogbb, XmNdefaultButton, dialogButton[dfltbutton - 1], NULL);
+    XtManageChild(dialogbb);
+
+    app = XtWidgetToApplicationContext(dialogbb);
+
+    dialogStatus = -1;
+
+    while (1)
+    {
+	XtAppNextEvent(app, &event);
+	XtDispatchEvent(&event);
+	if (dialogStatus >= 0)
+	    break;
+    }
+
+    XtUnmanageChild(dialogbb);
+
+    while (butcount)
+    {
+	--butcount;
+	XtUnmanageChild(dialogButton[butcount]);
+	XtDestroyWidget(dialogButton[butcount]);
+    }
+
+    return dialogStatus;
 }
+#endif /* GUI_DIALOG */
+
+/* ARGSUSED */
+    void
+gui_mch_show_popupmenu(menu)
+    GuiMenu *menu;
+{
+#ifdef MOTIF_POPUP
+    int		rootx, rooty, winx, winy;
+    Window	root, child;
+    unsigned int mask;
+
+    /* Position the popup menu at the pointer */
+    if (XQueryPointer(gui.dpy, XtWindow(vimShell), &root, &child,
+		&rootx, &rooty, &winx, &winy, &mask))
+    {
+	rootx -= 30;
+	if (rootx < 0)
+	    rootx = 0;
+	rooty -= 5;
+	if (rooty < 0)
+	    rooty = 0;
+	XtVaSetValues(menu->id,
+		XmNx, rootx,
+		XmNy, rooty,
+		NULL);
+    }
+
+    /* This causes a lockup.  How do we popup the menu? */
+    XtPopupSpringLoaded(menu->id);
 #endif
+}
