@@ -10,7 +10,7 @@
 
 #include "vim.h"
 
-#ifdef USE_CSCOPE
+#if defined(USE_CSCOPE) || defined(PROTO)
 
 #include <string.h>
 #include <errno.h>
@@ -37,7 +37,7 @@ static int	    cs_cnt_matches __ARGS((int idx));
 static char *	    cs_create_cmd __ARGS((char *csoption, char *pattern));
 static int	    cs_create_connection __ARGS((int i));
 static int	    cs_find __ARGS((EXARG *eap));
-static int	    cs_find_common __ARGS((char *opt, char *pat, int forceit));
+static int	    cs_find_common __ARGS((char *opt, char *pat, int, int ));
 static int	    cs_help __ARGS((EXARG *eap));
 static void	    cs_init __ARGS((void));
 static int	    cs_insert_filelist __ARGS((char *, char *, char *));
@@ -45,9 +45,10 @@ static int	    cs_kill __ARGS((EXARG *eap));
 static cscmd_t *    cs_lookup_cmd __ARGS((EXARG *eap));
 static char *	    cs_make_vim_style_matches __ARGS((char *, char *,
 			char *, char *));
-static char *	    cs_manage_matches __ARGS((char **, int , mcmd_e cmd));
-static char **	    cs_parse_results __ARGS((char *, int , int *, int *));
-static void	    cs_print_tags_priv __ARGS((char **, int ));
+static char *	    cs_manage_matches __ARGS((char **, char **, int, mcmd_e));
+static void	    cs_parse_results __ARGS((char *, int , int *, char ***,
+			char ***, int *));
+static void	    cs_print_tags_priv __ARGS((char **, char **, int));
 static int	    cs_read_prompt __ARGS((int ));
 static void	    cs_release_csp __ARGS((int, int freefnpp));
 static int	    cs_reset __ARGS((EXARG *eap));
@@ -120,47 +121,47 @@ do_cstag(eap)
     switch (p_csto) {
     case 0 :
 	if (cs_check_for_connections()) {
-	    ret = cs_find_common("g", (char *)(eap->arg), eap->forceit);
+	    ret = cs_find_common("g", (char *)(eap->arg), eap->forceit, FALSE);
 	    if (ret == FALSE) {
 		cs_free_tags();
 		if (msg_col)
 		    msg_putchar('\n');
 
 		if (cs_check_for_tags()) {
-		    (void)do_tag(eap->arg, DT_JUMP, 0, eap->forceit);
+		    ret = do_tag(eap->arg, DT_JUMP, 0, eap->forceit, FALSE);
 		}
 	    }
 	} else if (cs_check_for_tags()) {
-	    (void)do_tag(eap->arg, DT_JUMP, 0, eap->forceit);
-	} else {
-	    (void)EMSG("tag not found");
+	    ret = do_tag(eap->arg, DT_JUMP, 0, eap->forceit, FALSE);
 	}
 	break;
     case 1 :
 	if (cs_check_for_tags()) {
-	    ret = do_tag(eap->arg, DT_JUMP, 0, eap->forceit);
+	    ret = do_tag(eap->arg, DT_JUMP, 0, eap->forceit, FALSE);
 	    if (ret == FALSE) {
 		if (msg_col)
 		    msg_putchar('\n');
 
 		if (cs_check_for_connections()) {
-		    ret = cs_find_common("g", (char *)(eap->arg), eap->forceit);
+		    ret = cs_find_common("g", (char *)(eap->arg), eap->forceit,
+			    FALSE);
 		    if (ret == FALSE)
 			cs_free_tags();
 		}
 	    }
 	} else if (cs_check_for_connections()) {
-	    ret = cs_find_common("g", (char *)(eap->arg), eap->forceit);
+	    ret = cs_find_common("g", (char *)(eap->arg), eap->forceit, FALSE);
 	    if (ret == FALSE)
 		cs_free_tags();
-	} else {
-	    (void)EMSG("tag not found");
 	}
 	break;
     default :
 	break;
     }
 
+    if (!ret)
+	(void)EMSG("cstag: tag not found");
+
 } /* do_cscope */
 
 
@@ -172,14 +173,14 @@ do_cstag(eap)
  *
  * returns TRUE if eof, FALSE otherwise
  */
-int
+    int
 cs_fgets(buf, size)
     char_u	*buf;
     int		size;
 {
     char *p;
 
-    if ((p = cs_manage_matches(NULL, -1, Get)) == NULL)
+    if ((p = cs_manage_matches(NULL, NULL, -1, Get)) == NULL)
 	return TRUE;
 
     if (strlen(p) > size) {
@@ -198,11 +199,11 @@ cs_fgets(buf, size)
  *
  * called only from do_tag(), when popping the tag stack
  */
-void
+    void
 cs_free_tags()
 {
-    cs_manage_matches(NULL, -1, Free);
-} /* cs_free_tags */
+    cs_manage_matches(NULL, NULL, -1, Free);
+}
 
 
 /*
@@ -210,11 +211,11 @@ cs_free_tags()
  *
  * called from do_tag()
  */
-void
+    void
 cs_print_tags()
 {
-    cs_manage_matches(NULL, -1, Print);
-} /* cs_print_tags */
+    cs_manage_matches(NULL, NULL, -1, Print);
+}
 
 
 /*
@@ -229,13 +230,13 @@ cs_print_tags()
  *
  * MAXPATHL 256
  */
-static int
+/* ARGSUSED */
+    static int
 cs_add(eap)
     EXARG *eap;
 {
     char *fname, *ppath, *flags = NULL;
 
-    assert(strncmp("add", (const char *)(eap->arg), 3) == 0);
     if ((fname = strtok((char *)NULL, (const char *)" ")) == NULL) {
 	CS_USAGE_MSG(Add);
 	return CSCOPE_FAILURE;
@@ -244,88 +245,103 @@ cs_add(eap)
 	flags = strtok((char *)NULL, (const char *)" ");
 
     return cs_add_common(fname, ppath, flags);
-} /* cs_add */
+}
 
 
 /*
  * PRIVATE: cs_add_common
  *
  * the common routine to add a new cscope connection.  called by
- * cs_add() and cs_reset()
+ * cs_add() and cs_reset().  i really don't like to do this, but this
+ * routine uses a number of goto statements.
  */
-static int
-cs_add_common(fname, ppath, flags)
-    char *fname;
-    char *ppath;
+    static int
+cs_add_common(arg1, arg2, flags)
+    char *arg1;	    /* filename - may contain environment variables */
+    char *arg2;	    /* prepend path - may contain environment variables */
     char *flags;
 {
     struct stat statbuf;
     int ret;
-    char buf[MAXPATHL+64];
+    char *fname = NULL;
+    char *ppath = NULL;
+    char *buf = NULL;
     int i;
 
+    /* buf is for general purpose msgs.  let's hope MAXPATHL*2 is big enough */
+    if ((buf = (char *)alloc(MAXPATHL * 2)) == NULL) {
+	(void)EMSG("cs_add_common: alloc fail #1");
+	goto add_err;
+    }
+
+    /* get the filename (arg1), expand it, and try to stat it */
+    if ((fname = (char *)alloc(MAXPATHL+1)) == NULL) {
+	(void)EMSG("cs_add_common: alloc fail #2");
+	goto add_err;
+    }
+
+    expand_env((char_u *)arg1, (char_u *)fname, MAXPATHL);
     ret = stat(fname, &statbuf);
     if (ret < 0) {
-	if (errno == ENOENT) {
-	    if (p_csverbose)
-		(void)sprintf(buf, "%s does not exist (err %d)", fname, errno);
-	} else {
-	    if (p_csverbose)
-		(void)sprintf(buf, "Could not stat %s (err %d)", fname, errno);
-	}
-	if (p_csverbose)
+	if (p_csverbose) {
+	    (void)sprintf(buf, "stat(%s) error: %d)", fname, errno);
 	    (void)EMSG(buf);
-	return CSCOPE_FAILURE;
+	}
+	goto add_err;
     }
 
-    if (ppath != NULL) {
+    /* get the prepend path (arg2), expand it, and try to stat it */
+    if (arg2 != NULL) {
 	struct stat statbuf2;
+
+	if ((ppath = (char *)alloc(MAXPATHL+1)) == NULL) {
+	    (void)EMSG("cs_add_common: alloc fail #3");
+	    goto add_err;
+	}
+
+	expand_env((char_u *)arg2, (char_u *)ppath, MAXPATHL);
 	ret = stat(ppath, &statbuf2);
 	if (ret < 0) {
-	    if (errno == ENOENT) {
-		if (p_csverbose)
-		    (void)sprintf(buf, "%s does not exist (err %d)",
-			ppath, errno);
-	    } else {
-		if (p_csverbose)
-		    (void)sprintf(buf, "Could not stat %s (err %d)",
-			ppath, errno);
-	    }
-	    if (p_csverbose)
+	    if (p_csverbose) {
+		(void)sprintf(buf, "stat(%s) error: %d", fname, errno);
 		(void)EMSG(buf);
-	    return CSCOPE_FAILURE;
+	    }
+	    goto add_err;
 	}
     }
 
+    /* if filename is a directory, append the cscope database name to it */
     if ((statbuf.st_mode & S_IFMT) == S_IFDIR) {
+	char *fname2;
+	fname2 = (char *)alloc(strlen(CSCOPE_DBFILE) + strlen(fname));
+	if (fname2 == NULL) {
+	    (void)EMSG("cs_add_common: alloc fail #4");
+	    goto add_err;
+	}
+
 	while (fname[strlen(fname)-1] == '/') {
 	    fname[strlen(fname)-1] = '\0';
 	    if (strlen(fname) == 0)
 		break;
 	}
 	if (fname[0] == '\0')
-	    (void)sprintf(buf, "/%s", CSCOPE_DBFILE);
+	    (void)sprintf(fname2, "/%s", CSCOPE_DBFILE);
 	else
-	    (void)sprintf(buf, "%s/%s", fname, CSCOPE_DBFILE);
+	    (void)sprintf(fname2, "%s/%s", fname, CSCOPE_DBFILE);
 
-	ret = stat(buf, &statbuf);
+	ret = stat(fname2, &statbuf);
 	if (ret < 0) {
-	    if (errno == ENOENT) {
-		if (p_csverbose)
-		    (void)sprintf(buf, "%s does not exist (err %d)",
-			buf, errno);
-	    } else {
-		if (p_csverbose)
-		    (void)sprintf(buf, "Could not stat %s (err %d)",
-			buf, errno);
-	    }
-	    if (p_csverbose)
+	    if (p_csverbose) {
+		(void)sprintf(buf, "stat(%s) error: %d", fname2, errno);
 		(void)EMSG(buf);
-	    return CSCOPE_FAILURE;
+	    }
+	    vim_free(fname2);
+	    goto add_err;
 	}
-	i = cs_insert_filelist(buf, ppath, flags);
+	i = cs_insert_filelist(fname2, ppath, flags);
 	if (p_csverbose)
-	    (void)sprintf(buf, "Added cscope database %s", buf);
+	    (void)sprintf(buf, "Added cscope database %s", fname2);
+	vim_free(fname2);
     } else if ((statbuf.st_mode & S_IFMT) == S_IFREG) {
 	i = cs_insert_filelist(fname, ppath, flags);
 	if (p_csverbose)
@@ -336,18 +352,28 @@ cs_add_common(fname, ppath, flags)
 		"%s is not a directory or a valid cscope database", fname);
 	    (void)EMSG(buf);
 	}
-	return CSCOPE_FAILURE;
+	goto add_err;
     }
 
     if (i != -1) {
-        (void)cs_create_connection(i);
+	(void)cs_create_connection(i);
 	(void)cs_read_prompt(i);
 	if (p_csverbose) {
 	    msg_clr_eos();
 	    MSG_PUTS_ATTR(buf, highlight_attr[HLF_R]);
 	}
     }
+
+    vim_free(fname);
+    vim_free(ppath);
+    vim_free(buf);
     return CSCOPE_SUCCESS;
+
+add_err:
+    vim_free(fname);
+    vim_free(ppath);
+    vim_free(buf);
+    return CSCOPE_FAILURE;
 } /* cs_add_common */
 
 
@@ -500,17 +526,18 @@ static int
 cs_create_connection(i)
     int i;
 {
-    int to_cs[2], from_cs[2];
-    char cmd[MAXPATHLEN * 2];
+    int to_cs[2], from_cs[2], len;
+    char *prog, *cmd, *ppath = NULL;
 
     /*
      * Cscope reads from to_cs[0] and writes to from_cs[1]; vi reads from
      * from_cs[0] and writes to to_cs[1].
      */
-    to_cs[0] = to_cs[1] = from_cs[0] = from_cs[0] = -1;
+    to_cs[0] = to_cs[1] = from_cs[0] = from_cs[1] = -1;
     if (pipe(to_cs) < 0 || pipe(from_cs) < 0) {
 	goto err;
     }
+
     switch (csinfo[i].pid = fork()) {
     case -1:
 err:
@@ -525,16 +552,48 @@ err:
 	(void)EMSG("Could not create cscope pipes");
 	return CSCOPE_FAILURE;
     case 0:				/* child: run cscope. */
-	(void)dup2(to_cs[0], STDIN_FILENO);
-	(void)dup2(from_cs[1], STDOUT_FILENO);
-	(void)dup2(from_cs[1], STDERR_FILENO);
+	if (dup2(to_cs[0], STDIN_FILENO) == -1)
+	    perror("cs_create_connection 1");
+	if (dup2(from_cs[1], STDOUT_FILENO) == -1)
+	    perror("cs_create_connection 2");
+	if (dup2(from_cs[1], STDERR_FILENO) == -1)
+	    perror("cs_create_connection 3");
 
 	/* close unused */
 	(void)close(to_cs[1]);
 	(void)close(from_cs[0]);
 
+	/* expand the cscope exec for env var's */
+	if ((prog = (char *)alloc(MAXPATHL + 1)) == NULL) {
+	    perror("cs_create_connection: alloc fail #1");
+	    return CSCOPE_FAILURE;
+	}
+	expand_env((char_u *)p_csprg, (char_u *)prog, MAXPATHL);
+
+	/* alloc space to hold the cscope command */
+	len = strlen(prog) + strlen(csinfo[i].fname) + 32;
+	if (csinfo[i].ppath) {
+	    /* expand the prepend path for env var's */
+	    if ((ppath = (char *)alloc(MAXPATHL + 1)) == NULL) {
+		perror("cs_create_connection: alloc fail #2");
+		vim_free(prog);
+		return CSCOPE_FAILURE;
+	    }
+	    expand_env((char_u *)csinfo[i].ppath, (char_u *)ppath, MAXPATHL);
+
+	    len += strlen(ppath);
+	}
+	if (csinfo[i].flags)
+	    len += strlen(csinfo[i].flags);
+	if ((cmd = (char *)alloc(len)) == NULL) {
+	    perror("cs_create_connection: alloc fail #3");
+	    vim_free(prog);
+	    vim_free(ppath);
+	    return CSCOPE_FAILURE;
+	}
+
 	/* run the cscope command; is there execl for non-unix systems? */
-	(void)sprintf(cmd, "exec %s -dl -f %s", p_csexec, csinfo[i].fname);
+	(void)sprintf(cmd, "exec %s -dl -f %s", prog, csinfo[i].fname);
 	if (csinfo[i].ppath != NULL) {
 	    (void)strcat(cmd, " -P");
 	    (void)strcat(cmd, csinfo[i].ppath);
@@ -543,22 +602,28 @@ err:
 	    (void)strcat(cmd, " ");
 	    (void)strcat(cmd, csinfo[i].flags);
 	}
+	vim_free(prog);
+	vim_free(ppath);
 
-	(void)execl((const char *)p_sh, "sh", "-c", cmd, NULL);
+	if (execl("/bin/sh", "sh", "-c", cmd, NULL) == -1)
+	    perror("cs_create_connection exec failed");
 
 	exit (127);
 	/* NOTREACHED */
     default:	/* parent. */
-	/* close unused */
-	(void)close(to_cs[0]);
-	(void)close(from_cs[1]);
-
 	/*
 	 * Save the file descriptors for later duplication, and
 	 * reopen as streams.
 	 */
-	csinfo[i].to_fp = fdopen(to_cs[1], "w");
-	csinfo[i].fr_fp = fdopen(from_cs[0], "r");
+	if ((csinfo[i].to_fp = fdopen(to_cs[1], "w")) == NULL)
+	    perror("cs_create_connection: fdopen for to_fp failed");
+	if ((csinfo[i].fr_fp = fdopen(from_cs[0], "r")) == NULL)
+	    perror("cs_create_connection: fdopen for fr_fp failed");
+
+	/* close unused */
+	(void)close(to_cs[0]);
+	(void)close(from_cs[1]);
+
 	break;
     }
     return CSCOPE_SUCCESS;
@@ -580,8 +645,6 @@ cs_find(eap)
 {
     char *opt, *pat;
 
-    assert(strncmp("find", (const char *)(eap->arg), 4) == 0);
-
     if (cs_check_for_connections() == FALSE) {
 	(void)EMSG("no cscope connections");
 	return FALSE;
@@ -597,7 +660,7 @@ cs_find(eap)
 	return FALSE;
     }
 
-    return cs_find_common(opt, pat, eap->forceit);
+    return cs_find_common(opt, pat, eap->forceit, TRUE);
 } /* cs_find */
 
 
@@ -607,14 +670,15 @@ cs_find(eap)
  * common code for cscope find, shared by cs_find() and do_cstag()
  */
 static int
-cs_find_common(opt, pat, forceit)
+cs_find_common(opt, pat, forceit, verbose)
     char *opt;
     char *pat;
     int forceit;
+    int verbose;
 {
     int i;
     char *cmd;
-    char **matches;
+    char **matches, **contexts;
     int nummatches[CSCOPE_MAX_CONNECTIONS], totmatches, matched;
 
     /* create the actual command to send to cscope */
@@ -644,12 +708,15 @@ cs_find_common(opt, pat, forceit)
 	if (nummatches[i] == 0)
 	    (void)cs_read_prompt(i);
     }
-    free(cmd);
+    vim_free(cmd);
 
     if (totmatches == 0) {
 	char *nf = "no matches found in cscope connections";
 	char *buf;
 
+	if (!verbose)
+	    return FALSE;
+
 	buf = (char *) alloc(strlen(opt) + strlen(pat) + strlen(nf) + 1);
 	if (buf == NULL)
 	    (void)EMSG(nf);
@@ -657,19 +724,20 @@ cs_find_common(opt, pat, forceit)
 	    sprintf(buf, "no matches found for cscope query %s of %s",
 		opt, pat);
 	    (void)EMSG(buf);
-	    free(buf);
+	    vim_free(buf);
 	}
 	return FALSE;
     }
 
     /* read output */
-    matches = cs_parse_results((char *)pat, totmatches, nummatches, &matched);
+    cs_parse_results((char *)pat, totmatches, nummatches,
+	&matches, &contexts, &matched);
     if (matches == NULL)
 	return FALSE;
 
-    (void)cs_manage_matches(matches, totmatches, Store);
+    (void)cs_manage_matches(matches, contexts, totmatches, Store);
 
-    return do_tag((char_u *)pat, DT_CSCOPE, 0, forceit);
+    return do_tag((char_u *)pat, DT_CSCOPE, 0, forceit, verbose);
 
 } /* cs_find_common */
 
@@ -679,7 +747,8 @@ cs_find_common(opt, pat, forceit)
  *
  * print help
  */
-static int
+/* ARGSUSED */
+    static int
 cs_help(eap)
     EXARG *eap;
 {
@@ -757,7 +826,8 @@ cs_insert_filelist(fname, ppath, flags)
 
     if (ppath != NULL) {
 	if ((csinfo[i].ppath = (char *) alloc(strlen(ppath) + 1)) == NULL) {
-	    free(csinfo[i].fname);
+	    vim_free(csinfo[i].fname);
+	    csinfo[i].fname = NULL;
 	    return -1;
 	}
 	(void)strcpy(csinfo[i].ppath, (const char *)ppath);
@@ -766,8 +836,10 @@ cs_insert_filelist(fname, ppath, flags)
 
     if (flags != NULL) {
 	if ((csinfo[i].flags = (char *) alloc(strlen(flags) + 1)) == NULL) {
-	    free(csinfo[i].fname);
-	    free(csinfo[i].ppath);
+	    vim_free(csinfo[i].fname);
+	    vim_free(csinfo[i].ppath);
+	    csinfo[i].fname = NULL;
+	    csinfo[i].ppath = NULL;
 	    return -1;
 	}
 	(void)strcpy(csinfo[i].flags, (const char *)flags);
@@ -810,6 +882,7 @@ cs_lookup_cmd(exp)
  *
  * nuke em
  */
+/* ARGSUSED */
 static int
 cs_kill(eap)
     EXARG *eap;
@@ -817,7 +890,6 @@ cs_kill(eap)
     char *stok, printbuf[64];
     short i;
 
-    assert(strncmp("kill", (const char *)(eap->arg), 4) == 0);
     if ((stok = strtok((char *)NULL, (const char *)" ")) == NULL) {
 	CS_USAGE_MSG(Find);
 	return CSCOPE_FAILURE;
@@ -832,7 +904,7 @@ cs_kill(eap)
 	cs_release_csp(i, TRUE);
 	(void)sprintf(printbuf, "cscope connection %s closed", stok);
 	msg_clr_eos();
-        MSG_PUTS_ATTR(printbuf, highlight_attr[HLF_R]);
+	MSG_PUTS_ATTR(printbuf, highlight_attr[HLF_R]);
     }
 
     return 0;
@@ -923,12 +995,14 @@ cs_make_vim_style_matches(fname, slno, search, tagstr)
  * Print: prints the tags
  */
 static char *
-cs_manage_matches(matches, totmatches, cmd)
+cs_manage_matches(matches, contexts, totmatches, cmd)
     char **matches;
+    char **contexts;
     int totmatches;
     mcmd_e cmd;
 {
     static char **mp = NULL;
+    static char **cp = NULL;
     static int cnt = -1;
     static int next = -1;
     char *p = NULL;
@@ -937,9 +1011,10 @@ cs_manage_matches(matches, totmatches, cmd)
     case Store:
 	assert(matches != NULL);
 	assert(totmatches > 0);
-	if (mp != NULL)
-	    (void)cs_manage_matches(NULL, -1, Free);
+	if (mp != NULL || cp != NULL)
+	    (void)cs_manage_matches(NULL, NULL, -1, Free);
 	mp = matches;
+	cp = contexts;
 	cnt = totmatches;
 	next = 0;
 	break;
@@ -953,16 +1028,21 @@ cs_manage_matches(matches, totmatches, cmd)
     case Free:
 	if (mp != NULL) {
 	    if (cnt > 0)
-		while (cnt--)
-		    free(mp[cnt]);
-	    free(mp);
+		while (cnt--) {
+		    vim_free(mp[cnt]);
+		    if (cp != NULL)
+			vim_free(cp[cnt]);
+		}
+	    vim_free(mp);
+	    vim_free(cp);
 	}
 	mp = NULL;
+	cp = NULL;
 	cnt = 0;
 	next = 0;
 	break;
     case Print:
-	cs_print_tags_priv(mp, cnt);
+	cs_print_tags_priv(mp, cp, cnt);
 	break;
     default:	/* should not reach here */
 	(void)EMSG("fatal error in cs_manage_matches");
@@ -979,27 +1059,32 @@ cs_manage_matches(matches, totmatches, cmd)
  * parse cscope output and calls cs_make_vim_style_matches to convert
  * into ctags format
  */
-static char **
-cs_parse_results(tagstr, totmatches, nummatches_a, matched)
+static void
+cs_parse_results(tagstr, totmatches, nummatches_a, matches_p, cntxts_p, matched)
     char *tagstr;
     int totmatches;
     int *nummatches_a;
+    char ***matches_p;
+    char ***cntxts_p;
     int *matched;
 {
     int i, j;
     int ch;
     char buf[2048];
     char *name, *search, *p, *slno;
-    int totsofar;
-    char **matches;
+    int totsofar = 0;
+    char **matches = NULL;
+    char **cntxts = NULL;
     char *fullname;
+    char *cntx;
 
     assert(totmatches > 0);
 
     if ((matches = (char **) alloc(sizeof(char *) * totmatches)) == NULL)
-	return NULL;
+	goto parse_out;
+    if ((cntxts = (char **) alloc(sizeof(char *) * totmatches)) == NULL)
+	goto parse_out;
 
-    totsofar = 0;
     for (i = 0; i < CSCOPE_MAX_CONNECTIONS; i++) {
 	if (nummatches_a[i] < 1)
 	    continue;
@@ -1011,7 +1096,7 @@ cs_parse_results(tagstr, totmatches, nummatches_a, matched)
 		    errno = EIO;
 		(void)sprintf(buf, "error reading cscope connection %d", i);
 		(void)EMSG(buf);
-		return NULL;
+		goto parse_out;
 	    }
 
 	    /* If the line's too long for the buffer, discard it. */
@@ -1029,7 +1114,7 @@ cs_parse_results(tagstr, totmatches, nummatches_a, matched)
 	     */
 	    if ((name = strtok((char *)buf, (const char *)" ")) == NULL)
 		continue;
-	    if ((slno = strtok(NULL, (const char *)" ")) == NULL)
+	    if ((cntx = strtok(NULL, (const char *)" ")) == NULL)
 		continue;
 	    if ((slno = strtok(NULL, (const char *)" ")) == NULL)
 		continue;
@@ -1050,20 +1135,29 @@ cs_parse_results(tagstr, totmatches, nummatches_a, matched)
 	    matches[totsofar] =
 		cs_make_vim_style_matches(fullname, slno, search, tagstr);
 
-	    free(fullname);
+	    vim_free(fullname);
+
+	    if (strcmp(cntx, "<global>") == 0)
+		cntxts[totsofar] = NULL;
+	    else
+		/* note: if vim_strsave returns NULL, then the context
+		 * will be "<global>", which is misleading.
+		 */
+		cntxts[totsofar] = (char *)vim_strsave((char_u *)cntx);
 
 	    if (matches[totsofar] != NULL)
 		totsofar++;
 
 	} /* for all matches */
 
-        (void)cs_read_prompt(i);
+	(void)cs_read_prompt(i);
 
     } /* for all cscope connections */
 
+parse_out:
     *matched = totsofar;
-
-    return matches;
+    *matches_p = matches;
+    *cntxts_p = cntxts;
 } /* cs_parse_results */
 
 
@@ -1073,8 +1167,9 @@ cs_parse_results(tagstr, totmatches, nummatches_a, matched)
  * called from cs_manage_matches()
  */
 static void
-cs_print_tags_priv(matches, num_matches)
+cs_print_tags_priv(matches, cntxts, num_matches)
     char **matches;
+    char **cntxts;
     int num_matches;
 {
     char buf[1024]; /* hope that 1024 is enough, else we need to malloc */
@@ -1088,25 +1183,25 @@ cs_print_tags_priv(matches, num_matches)
     (void)sprintf(buf, "Cscope tag: %s\n", tname);
     MSG_PUTS_ATTR(buf, highlight_attr[HLF_T]);
 
-    MSG_PUTS_ATTR("  #   line", highlight_attr[HLF_T]);	/* strlen is 7 */
+    MSG_PUTS_ATTR("   #   line", highlight_attr[HLF_T]);    /* strlen is 7 */
     msg_advance(msg_col + 2);
-    MSG_PUTS_ATTR("filename\n", highlight_attr[HLF_T]);
+    MSG_PUTS_ATTR("filename / context / line\n", highlight_attr[HLF_T]);
 
     for (i = 0; i < num_matches; i++) {
 	/* if we really wanted to, we could avoid this malloc and strcpy
 	 * by parsing matches[i] on the fly and placing stuff into buf
 	 * directly, but that's too much of a hassle
 	 */
-	if ((tbuf = (char *) alloc(strlen(matches[i]) + 1)) == NULL) {
+	if ((tbuf = (char *)alloc(strlen(matches[i]) + 1)) == NULL)
 	    continue;
-	}
 	(void)strcpy(tbuf, matches[i]);
 
 	if ((fname = strtok(tbuf, (const char *)"\t")) == NULL)
 	    continue;
 	if ((fname = strtok(NULL, (const char *)"\t")) == NULL)
 	    continue;
-	if ((lno = strtok(NULL, (const char *)"\t")) == NULL) {
+	if ((lno = strtok(NULL, (const char *)"\t")) == NULL)
+	{
 	    /* if NULL, then no "extra", although in cscope's case, there
 	     * should always be "extra".
 	     */
@@ -1117,17 +1212,27 @@ cs_print_tags_priv(matches, num_matches)
 
 	lno[strlen(lno)-2] = '\0';  /* ignore ;" at the end */
 
-	(void)sprintf(buf, " %2d %6s  %s", i + 1, lno, fname);
-	MSG_PUTS_ATTR(buf, highlight_attr[HLF_D]);
-	if (msg_col)
-	    msg_putchar('\n');
+	(void)sprintf(buf, "%4d %6s  ", i + 1, lno);
+	MSG_PUTS_ATTR(buf, highlight_attr[HLF_CM]);
+	MSG_PUTS_LONG_ATTR(fname, highlight_attr[HLF_CM]);
 
-	if (extra != NULL) {
-	    msg_advance(12);
-	    MSG_PUTS(extra);
+	(void)sprintf(buf, " <<%s>>",
+		(cntxts[i] == NULL) ? "GLOBAL" : cntxts[i]);
+
+	/* print the context only if it fits on the same line */
+	if (msg_col + strlen(buf) >= (int)Columns)
+	    msg_putchar('\n');
+	msg_advance(12);
+	MSG_PUTS_LONG(buf);
+	msg_putchar('\n');
+
+	if (extra != NULL)
+	{
+	    msg_advance(13);
+	    MSG_PUTS_LONG(extra);
 	}
 
-	free(tbuf);
+	vim_free(tbuf); /* only free after printing extra due to strtok use */
 
 	if (msg_col)
 	    msg_putchar('\n');
@@ -1157,12 +1262,21 @@ cs_read_prompt(i)
 	while ((ch = getc(csinfo[i].fr_fp)) != EOF && ch != CSCOPE_PROMPT[0])
 	    ;
 	if (ch == EOF) {
+	    perror("cs_read_prompt EOF(1)");
 	    cs_release_csp(i, TRUE);
 	    return CSCOPE_FAILURE;
 	}
-	if (getc(csinfo[i].fr_fp) != CSCOPE_PROMPT[1])
+
+	ch = getc(csinfo[i].fr_fp);
+	if (ch == EOF)
+	    perror("cs_read_prompt EOF(2)");
+	if (ch != CSCOPE_PROMPT[1])
 	    continue;
-	if (getc(csinfo[i].fr_fp) != CSCOPE_PROMPT[2])
+
+	ch = getc(csinfo[i].fr_fp);
+	if (ch == EOF)
+	    perror("cs_read_prompt EOF(3)");
+	if (ch != CSCOPE_PROMPT[2])
 	    continue;
 	break;
     }
@@ -1197,7 +1311,7 @@ cs_release_csp(i, freefnpp)
     (void)waitpid(tpid, &pstat, WNOHANG);
 
     if (freefnpp) {
-        vim_free(csinfo[i].fname);
+	vim_free(csinfo[i].fname);
 	vim_free(csinfo[i].ppath);
 	vim_free(csinfo[i].flags);
     }
@@ -1216,6 +1330,7 @@ cs_release_csp(i, freefnpp)
  *
  * calls cs_kill on all cscope connections then reinits
  */
+/* ARGSUSED */
 static int
 cs_reset(eap)
     EXARG *eap;
@@ -1224,8 +1339,6 @@ cs_reset(eap)
     int	i;
     char	buf[8];
 
-    assert(strncmp("reset", (const char *)(eap->arg), 5) == 0);
-
     /* malloc our db and ppath list */
     dblist = (char **) alloc(CSCOPE_MAX_CONNECTIONS * sizeof(char *));
     pplist = (char **) alloc(CSCOPE_MAX_CONNECTIONS * sizeof(char *));
@@ -1238,7 +1351,7 @@ cs_reset(eap)
     }
 
     for (i = 0; i < CSCOPE_MAX_CONNECTIONS; i++) {
-        dblist[i] = csinfo[i].fname;
+	dblist[i] = csinfo[i].fname;
 	pplist[i] = csinfo[i].ppath;
 	fllist[i] = csinfo[i].flags;
 	if (csinfo[i].fname != NULL)
@@ -1256,9 +1369,9 @@ cs_reset(eap)
 	vim_free(pplist[i]);
 	vim_free(fllist[i]);
     }
-    free(dblist);
-    free(pplist);
-    free(fllist);
+    vim_free(dblist);
+    vim_free(pplist);
+    vim_free(fllist);
 
     MSG_PUTS_ATTR("All cscope databases reset", highlight_attr[HLF_R]);
     return CSCOPE_SUCCESS;
@@ -1292,9 +1405,9 @@ cs_resolve_file(i, name)
      */
     len = strlen(name) + 2;
     if (csinfo[i].ppath != NULL)
-        len += strlen(csinfo[i].ppath);
+	len += strlen(csinfo[i].ppath);
 
-    if ((fullname = (char *) alloc(len)) == NULL)
+    if ((fullname = (char *)alloc(len)) == NULL)
 	return NULL;
 
     /*
@@ -1303,10 +1416,11 @@ cs_resolve_file(i, name)
      * happens, you are screwed up and need to fix how you're using cscope.
      */
     if (csinfo[i].ppath != NULL &&
-	(strncmp(name, csinfo[i].ppath, strlen(csinfo[i].ppath)) != 0))
-        (void)sprintf(fullname, "%s/%s", csinfo[i].ppath, name);
+	(strncmp(name, csinfo[i].ppath, strlen(csinfo[i].ppath)) != 0) &&
+	(name[0] != '/'))
+	(void)sprintf(fullname, "%s/%s", csinfo[i].ppath, name);
     else
-        (void)sprintf(fullname, "%s", name);	/* hmm, do we need a "./" ? */
+	(void)sprintf(fullname, "%s", name);
 
     return fullname;
 } /* cs_resolve_file */
@@ -1317,6 +1431,7 @@ cs_resolve_file(i, name)
  *
  * show all cscope connections
  */
+/* ARGSUSED */
 static int
 cs_show(eap)
     EXARG *eap;
@@ -1324,8 +1439,6 @@ cs_show(eap)
     char buf[MAXPATHL*2];
     short i;
 
-    assert(strncmp("show", (const char *)(eap->arg), 4) == 0);
-
     if (cs_cnt_connections() == 0) {
 	MSG_PUTS("no cscope connections\n");
     } else {
@@ -1337,11 +1450,11 @@ cs_show(eap)
 		continue;
 
 	    if (csinfo[i].ppath != NULL)
-		(void)sprintf(buf, "%2d %-5d  %-34s  %-32s\n",
-		    i, csinfo[i].pid, csinfo[i].fname, csinfo[i].ppath);
+		(void)sprintf(buf, "%2d %-5ld  %-34s  %-32s\n",
+		    i, (long)csinfo[i].pid, csinfo[i].fname, csinfo[i].ppath);
 	    else
-		(void)sprintf(buf, "%2d %-5d  %-34s  <none>\n",
-		    i, csinfo[i].pid, csinfo[i].fname);
+		(void)sprintf(buf, "%2d %-5ld  %-34s  <none>\n",
+		    i, (long)csinfo[i].pid, csinfo[i].fname);
 	    MSG_PUTS(buf);
 	}
     }
